from fbs.freeze.linux import freeze_linux, remove_shared_libraries

def freeze_fedora(debug=False):
    freeze_linux(debug)
    # Force Fedora to use the system's Gnome libraries. This avoids warnings
    # when starting the app on the command line.
    remove_shared_libraries('libgio-2.0.so.*', 'libglib-2.0.so.*')
    # Fixes for Fedora 29:
    remove_shared_libraries('libfreetype.so.*', 'libssl.so.*')
    # PyInstaller 3.4 includes the library below when on Python 3.6.
    # (Interestingly, it does not package it on Python 3.5.) This leads to a lot
    # of Fontconfig-related errors when starting the frozen app. Further,
    # starting the app takes ages. Removing the library fixes this:
    remove_shared_libraries('libfontconfig.so.*')


from fbs import path
from fbs._gpg import preset_gpg_passphrase
from fbs.resources import copy_with_filtering
from fbs_runtime._source import default_path
from os import makedirs
from os.path import exists
from shutil import rmtree
from subprocess import check_call, DEVNULL

def create_repo_ubuntu():
    dest_dir = path('target/repo')
    tmp_dir = path('target/repo-tmp')
    if exists(dest_dir):
        rmtree(dest_dir)
    if exists(tmp_dir):
        rmtree(tmp_dir)
    makedirs(tmp_dir)
    distr_file = 'src/repo/ubuntu/distributions'
    distr_path = path(distr_file)
    if not exists(distr_path):
        distr_path = default_path(distr_file)
    copy_with_filtering(distr_path, tmp_dir, files_to_filter=[distr_path])
    preset_gpg_passphrase()
    check_call([
        'reprepro', '-b', dest_dir, '--confdir', tmp_dir,
        'includedeb', 'stable', path('target/${installer}')
    ], stdout=DEVNULL)


from fbs import path, SETTINGS
from fbs_runtime import FbsError
from os import makedirs
from os.path import join, splitext, dirname, basename, exists
from shutil import copy
from subprocess import call, run, DEVNULL

import hashlib
import json
import os

_CERTIFICATE_PATH = 'src/sign/windows/certificate.pfx'
_TO_SIGN = ('.exe', '.cab', '.dll', '.ocx', '.msi', '.xpi')

def sign_windows():
    if not exists(path(_CERTIFICATE_PATH)):
        raise FbsError(
            'Could not find a code signing certificate at:\n    '
            + _CERTIFICATE_PATH
        )
    if 'windows_sign_pass' not in SETTINGS:
        raise FbsError(
            "Please set 'windows_sign_pass' to the password of %s in either "
            "src/build/settings/secret.json, .../windows.json or .../base.json."
            % _CERTIFICATE_PATH
        )
    for subdir, _, files in os.walk(path('${freeze_dir}')):
        for file_ in files:
            extension = splitext(file_)[1]
            if extension in _TO_SIGN:
                sign_file(join(subdir, file_))

def sign_file(file_path, description='', url=''):
    helper = _SignHelper.instance()
    if not helper.is_signed(file_path):
        helper.sign(file_path, description, url)

class _SignHelper:

    _INSTANCE = None

    @classmethod
    def instance(cls):
        if cls._INSTANCE is None:
            cls._INSTANCE = cls(path('cache/signed'))
        return cls._INSTANCE

    def __init__(self, cache_dir):
        self._cache_dir = cache_dir

    def is_signed(self, file_path):
        return not call(
            ['signtool', 'verify', '/pa', file_path], stdout=DEVNULL,
            stderr=DEVNULL
        )

    def sign(self, file_path, description, url):
        json_path = self._get_json_path(file_path)
        try:
            with open(json_path) as f:
                cached = json.load(f)
            is_in_cache = description == cached['description'] and \
                          url == cached['url'] and \
                          self._hash(file_path) == cached['hash']
        except FileNotFoundError:
            is_in_cache = False
        if not is_in_cache:
            self._sign(file_path, description, url)
        copy(self._get_path_in_cache(file_path), file_path)

    def _sign(self, file_path, description, url):
        path_in_cache = self._get_path_in_cache(file_path)
        makedirs(dirname(path_in_cache), exist_ok=True)
        copy(file_path, path_in_cache)
        hash_ = self._hash(path_in_cache)
        self._run_signtool(path_in_cache, 'sha1')
        self._run_signtool(path_in_cache, 'sha256')
        with open(self._get_json_path(file_path), 'w') as f:
            json.dump({
                'description': description,
                'url': url,
                'hash': hash_
            }, f)

    def _get_json_path(self, file_path):
        return self._get_path_in_cache(file_path) + '.json'

    def _get_path_in_cache(self, file_path):
        return join(self._cache_dir, basename(file_path))

    def _run_signtool(self, file_path, digest_alg, description='', url=''):
        password = SETTINGS['windows_sign_pass']
        args = [
            'signtool', 'sign', '/f', path(_CERTIFICATE_PATH), '/p', password
        ]
        if 'windows_sign_server' in SETTINGS:
            args.extend(['/tr', SETTINGS['windows_sign_server']])
        if description:
            args.extend(['/d', description])
        if url:
            args.extend(['/du', url])
        args.extend(['/as', '/fd', digest_alg, '/td', digest_alg])
        args.append(file_path)
        run(args, check=True, stdout=DEVNULL)

    def _hash(self, file_path):
        bufsize = 65536
        hasher = hashlib.md5()
        with open(file_path, 'rb') as f:
            buf = f.read(bufsize)
            while buf:
                hasher.update(buf)
                buf = f.read(bufsize)
        return hasher.hexdigest()


from fbs import path, SETTINGS
from fbs.sign.windows import sign_file

def sign_installer_windows():
    installer = path('target/${installer}')
    sign_file(installer, SETTINGS['app_name'] + ' Setup', SETTINGS['url'])


from fbs import path, SETTINGS
from fbs_runtime import FbsError
from os import makedirs
from os.path import exists, join
from shutil import rmtree, copy
from subprocess import check_call, DEVNULL

def create_repo_arch():
    if not exists(path('target/${installer}.sig')):
        raise FbsError(
            'Installer does not exist or is not signed. Maybe you need to '
            'run:\n'
            '    fbs signinst'
        )
    dest_dir = path('target/repo')
    if exists(dest_dir):
        rmtree(dest_dir)
    makedirs(dest_dir)
    app_name = SETTINGS['app_name']
    pkg_file = path('target/${installer}')
    pkg_file_versioned = '%s-%s.pkg.tar.xz' % (app_name, SETTINGS['version'])
    copy(pkg_file, join(dest_dir, pkg_file_versioned))
    copy(pkg_file + '.sig', join(dest_dir, pkg_file_versioned + '.sig'))
    check_call(
        ['repo-add', '%s.db.tar.gz' % app_name, pkg_file_versioned],
        cwd=dest_dir, stdout=DEVNULL
    )
    # Ensure the permissions are correct if uploading to a server:
    check_call(['chmod', 'g-w', '-R', dest_dir])


from fbs import path, SETTINGS
from fbs.freeze import run_pyinstaller, _generate_resources
from fbs.resources import _copy
from fbs_runtime._source import default_path
from os import remove
from os.path import join, exists
from shutil import copy

import os
import struct
import sys

def freeze_windows(debug=False):
    args = []
    if not (debug or SETTINGS['show_console_window']):
        # The --windowed flag below prevents us from seeing any console output.
        # We therefore only add it when we're not debugging.
        args.append('--windowed')
    args.extend(['--icon', path('src/main/icons/Icon.ico')])
    for path_fn in default_path, path:
        _copy(path_fn, 'src/freeze/windows/version_info.py', path('target/PyInstaller'))
    args.extend(['--version-file', path('target/PyInstaller/version_info.py')])
    run_pyinstaller(args, debug)
    _restore_corrupted_python_dlls()
    _generate_resources()
    copy(path('src/main/icons/Icon.ico'), path('${freeze_dir}'))
    _add_missing_dlls()

def _restore_corrupted_python_dlls():
    # PyInstaller <= 3.4 somehow corrupts python3*.dll - see:
    # https://github.com/pyinstaller/pyinstaller/issues/2526
    # Restore the uncorrupted original:
    python_dlls = (
        'python%s.dll' % sys.version_info.major,
        'python%s%s.dll' % (sys.version_info.major, sys.version_info.minor)
    )
    for dll_name in python_dlls:
        try:
            remove(path('${freeze_dir}/' + dll_name))
        except FileNotFoundError:
            pass
        else:
            copy(_find_on_path(dll_name), path('${freeze_dir}'))

def _add_missing_dlls():
    for dll_name in (
        'msvcr100.dll', 'msvcr110.dll', 'msvcp110.dll', 'vcruntime140.dll',
        'msvcp140.dll', 'concrt140.dll', 'vccorlib140.dll'
    ):
        try:
            _add_missing_dll(dll_name)
        except LookupError:
            raise FileNotFoundError(
                "Could not find %s on your PATH. Please install the Visual C++ "
                "Redistributable for Visual Studio 2012 from:\n    "
                "https://www.microsoft.com/en-us/download/details.aspx?id=30679"
                % dll_name
            ) from None
    for ucrt_dll in ('api-ms-win-crt-multibyte-l1-1-0.dll',):
        try:
            _add_missing_dll(ucrt_dll)
        except LookupError:
            bitness_32_or_64 = struct.calcsize("P") * 8
            raise FileNotFoundError(
                "Could not find %s on your PATH. If you are on Windows 10, you "
                "may have to install the Windows 10 SDK from "
                "https://developer.microsoft.com/en-us/windows/downloads/windows-10-sdk. "
                "Otherwise, try installing KB2999226 from "
                "https://support.microsoft.com/en-us/kb/2999226. "
                "In both cases, add the directory containing %s to your PATH "
                "environment variable afterwards. If there are 32 and 64 bit "
                "versions of the DLL, use the %s bit one (because that's the "
                "bitness of your current Python interpreter)." % (
                    ucrt_dll, ucrt_dll, bitness_32_or_64
                )
            ) from None

def _add_missing_dll(dll_name):
    freeze_dir = path('${freeze_dir}')
    if not exists(join(freeze_dir, dll_name)):
        copy(_find_on_path(dll_name), freeze_dir)

def _find_on_path(file_name):
    path = os.environ.get("PATH", os.defpath)
    path_items = path.split(os.pathsep)
    if sys.platform == "win32":
        if not os.curdir in path_items:
            path_items.insert(0, os.curdir)
    seen = set()
    for dir_ in path_items:
        normdir = os.path.normcase(dir_)
        if not normdir in seen:
            seen.add(normdir)
            file_path = join(dir_, file_name)
            if exists(file_path):
                return file_path
    raise LookupError("Could not find %s on PATH" % file_name)



"""
This hook is in particular required to prevent errors with the following code
under PySide2 5.12.1:

    from PySide2.QtWidgets import QApplication
    print(QApplication.__signature__)

It should print

    [<Signature (self)>, <Signature (self, arg__1:typing.List)>]

Without this hook, it prints a lot of output of the form

    parser.py:191: RuntimeWarning: pyside_type_init:
    UNRECOGNIZED: 'PySide2.QtGui....'

Then, it prints Signature('QStringList') instead of the above ...(List).
"""

from glob import glob
from os.path import dirname, relpath, join

import PySide2.support

"""
This should give roughly the same results as:
 
    from PyInstaller.utils.hooks import collect_data_files
    datas = collect_data_files(
        'PySide2', include_py_files=True, subdir='support'
    )

The reason we don't do it this way is that it would add a dynamic link to
PyInstaller, and thus force the GPL on fbs, preventing it from being licensed
under different terms (such as a commercial license).
"""
_base_dir = dirname(PySide2.support.__file__)
_python_files = glob(join(_base_dir, '**', '*.py'), recursive=True)
_site_packages = dirname(dirname(_base_dir))
datas = [(f, relpath(dirname(f), _site_packages)) for f in _python_files]


from fbs import path
from fbs.freeze import _generate_resources, run_pyinstaller
from glob import glob
from os import remove
from shutil import copy

def freeze_linux(debug=False):
    run_pyinstaller(debug=debug)
    _generate_resources()
    copy(path('src/main/icons/Icon.ico'), path('${freeze_dir}'))
    # For some reason, PyInstaller packages libstdc++.so.6 even though it is
    # available on most Linux distributions. If we include it and run our app on
    # a different Ubuntu version, then Popen(...) calls fail with errors
    # "GLIBCXX_... not found" or "CXXABI_..." not found. So ensure we don't
    # package the file, so that the respective system's compatible version is
    # used:
    remove_shared_libraries(
        'libstdc++.so.*', 'libtinfo.so.*', 'libreadline.so.*', 'libdrm.so.*'
    )

def remove_shared_libraries(*filename_patterns):
    for pattern in filename_patterns:
        for file_path in glob(path('${freeze_dir}/' + pattern)):
            remove(file_path)


from fbs import _server, SETTINGS, path
from fbs._aws import upload_file, upload_folder_contents
from fbs_runtime import FbsError
from fbs_runtime.platform import is_linux
from os.path import basename

import json

def _upload_repo(username, password):
    status, response = _server.post_json('start_upload', {
        'username': username,
        'password': password
    })
    unexpected_response = lambda: FbsError(
        'Received unexpected server response %d:\n%s' % (status, response)
    )
    if status // 2 != 100:
        raise unexpected_response()
    try:
        data = json.loads(response)
    except ValueError:
        raise unexpected_response()
    try:
        credentials = data['bucket'], data['key'], data['secret']
    except KeyError:
        raise unexpected_response()
    dest_path = lambda p: username + '/' + SETTINGS['app_name'] + '/' + p
    installer = path('target/${installer}')
    installer_dest = dest_path(basename(installer))
    upload_file(installer, installer_dest, *credentials)
    uploaded = [installer_dest]
    if is_linux():
        repo_dest = dest_path(SETTINGS['repo_subdir'])
        uploaded.extend(
            upload_folder_contents(path('target/repo'), repo_dest, *credentials)
        )
        pubkey_dest = dest_path('public-key.gpg')
        upload_file(
            path('src/sign/linux/public-key.gpg'), pubkey_dest, *credentials
        )
        uploaded.append(pubkey_dest)
    status, response = _server.post_json('complete_upload', {
        'username': username,
        'password': password,
        'files': uploaded
    })
    if status != 201:
        raise unexpected_response()


from fbs import path, SETTINGS
from fbs.builtin_commands._docker import _run_docker
from fbs.builtin_commands._util import prompt_for_value, update_json, \
    require_existing_project, BASE_JSON, SECRET_JSON
from fbs.cmdline import command
from fbs_runtime import FbsError
from os import makedirs
from os.path import dirname, exists
from pathlib import Path
from subprocess import DEVNULL, PIPE

import logging

_LOG = logging.getLogger(__name__)
_DOCKER_IMAGE = 'fbs/gpg-generator'
_DEST_DIR = 'src/sign/linux'
_PUBKEY_NAME = 'public-key.gpg'
_PRIVKEY_NAME = 'private-key.gpg'

@command
def gengpgkey():
    """
    Generate a GPG key for Linux code signing
    """
    require_existing_project()
    if exists(_DEST_DIR):
        raise FbsError('The %s folder already exists. Aborting.' % _DEST_DIR)
    try:
        email = prompt_for_value('Email address')
        name = prompt_for_value('Real name', default=SETTINGS['author'])
        passphrase = prompt_for_value('Key password', password=True)
    except KeyboardInterrupt:
        print('')
        return
    print('')
    _LOG.info('Generating the GPG key. This can take a little...')
    _init_docker()
    args = ['run', '-t']
    if exists('/dev/urandom'):
        # Give the key generator more entropy on Posix:
        args.extend(['-v', '/dev/urandom:/dev/random'])
    args.extend([_DOCKER_IMAGE, '/root/genkey.sh', name, email, passphrase])
    result = _run_docker(args, check=True, stdout=PIPE, universal_newlines=True)
    key = _snip(
        result.stdout,
        "revocation certificate stored as '/root/.gnupg/openpgp-revocs.d/",
        ".rev'",
        include_bounds=False
    )
    pubkey = _snip(result.stdout,
                   '-----BEGIN PGP PUBLIC KEY BLOCK-----\n',
                   '-----END PGP PUBLIC KEY BLOCK-----\n')
    privkey = _snip(result.stdout,
                    '-----BEGIN PGP PRIVATE KEY BLOCK-----\n',
                    '-----END PGP PRIVATE KEY BLOCK-----\n')
    makedirs(path(_DEST_DIR), exist_ok=True)
    pubkey_dest = _DEST_DIR + '/' + _PUBKEY_NAME
    Path(path(pubkey_dest)).write_text(pubkey)
    Path(path(_DEST_DIR + '/' + _PRIVKEY_NAME)).write_text(privkey)
    update_json(path(BASE_JSON), {'gpg_key': key, 'gpg_name': name})
    update_json(path(SECRET_JSON), {'gpg_pass': passphrase})
    _LOG.info(
        'Done. Created %s and ...%s. Also updated %s and ...secret.json with '
        'the values you provided.', pubkey_dest, _PRIVKEY_NAME, BASE_JSON
    )

def _init_docker():
    _run_docker(
        ['build', '-t', _DOCKER_IMAGE, dirname(__file__)], stdout=DEVNULL
    )

def _snip(str_, preamble, postamble, include_bounds=True):
    start = str_.index(preamble)
    end = str_.index(postamble, start + len(preamble))
    if not include_bounds:
        start += len(preamble)
    if include_bounds:
        end += len(postamble)
    return str_[start:end]


"""Create cross-platform desktop applications with Python and Qt

See:
https://build-system.fman.io
"""

from os.path import relpath, join
from setuptools import setup, find_packages

import os

def _get_package_data(pkg_dir, data_subdir):
    result = []
    for dirpath, _, filenames in os.walk(join(pkg_dir, data_subdir)):
        for filename in filenames:
            filepath = join(dirpath, filename)
            result.append(relpath(filepath, pkg_dir))
    return result

description = 'Create cross-platform desktop applications with Python and Qt'
setup(
    name='fbs',
    # Also update fbs/_defaults/requirements/base.txt when you change this:
    version='1.0.3',
    description=description,
    long_description=
        description + '\n\nHome page: https://build-system.fman.io',
    author='Michael Herrmann',
    author_email='michael+removethisifyouarehuman@herrmann.io',
    url='https://build-system.fman.io',
    packages=find_packages(exclude=('tests', 'tests.*')),
    package_data={
        'fbs': _get_package_data('fbs', '_defaults'),
        'fbs.builtin_commands':
            _get_package_data('fbs/builtin_commands', 'project_template'),
        'fbs.builtin_commands._gpg':
            ['Dockerfile', 'genkey.sh', 'gpg-agent.conf'],
        'fbs.installer.mac': _get_package_data(
            'fbs/installer/mac', 'create-dmg'
        )
    },
    install_requires=['PyInstaller==3.4'],
    extras_require={
        # Also update requirements.txt when you change this:
        'licensing': ['rsa>=3.4.2'],
        'sentry': ['sentry-sdk>=0.6.6'],
        'upload': ['boto3']
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
    
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
    
        'Operating System :: OS Independent',
    
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',

        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    entry_points={
        'console_scripts': ['fbs=fbs.__main__:_main']
    },
    license='GPLv3 or later',
    keywords='PyQt',
    platforms=['MacOS', 'Windows', 'Debian', 'Fedora', 'CentOS', 'Arch'],
    test_suite='tests'
)


from fbs import path, SETTINGS
from fbs.freeze import _generate_resources, run_pyinstaller
from fbs.resources import get_icons
from os import makedirs, unlink, rename, symlink
from os.path import exists
from shutil import copy, rmtree
from subprocess import run

def freeze_mac(debug=False):
    if not exists(path('target/Icon.icns')):
        _generate_iconset()
        run(['iconutil', '-c', 'icns', path('target/Icon.iconset')], check=True)
    args = []
    if not (debug or SETTINGS['show_console_window']):
        args.append('--windowed')
    args.extend(['--icon', path('target/Icon.icns')])
    bundle_identifier = SETTINGS['mac_bundle_identifier']
    if bundle_identifier:
        args.extend([
            '--osx-bundle-identifier', bundle_identifier
        ])
    run_pyinstaller(args, debug)
    _remove_unwanted_pyinstaller_files()
    _fix_sparkle_delta_updates()
    _generate_resources()

def _generate_iconset():
    makedirs(path('target/Icon.iconset'), exist_ok=True)
    for size, scale, icon_path in get_icons():
        dest_name = 'icon_%dx%d' % (size, size)
        if scale != 1:
            dest_name += '@%dx' % scale
        dest_name += '.png'
        copy(icon_path, path('target/Icon.iconset/' + dest_name))

def _remove_unwanted_pyinstaller_files():
    for unwanted in ('include', 'lib', 'lib2to3'):
        try:
            unlink(path('${freeze_dir}/Contents/MacOS/' + unwanted))
        except FileNotFoundError:
            pass
        try:
            rmtree(path('${freeze_dir}/Contents/Resources/' + unwanted))
        except FileNotFoundError:
            pass

def _fix_sparkle_delta_updates():
    # Sparkle's Delta Updates mechanism does not support signed non-Mach-O files
    # in Contents/MacOS. base_library.zip, which is created by PyInstaller,
    # violates this. We therefore move base_library.zip to Contents/Resources.
    # Fortunately, everything still works if we then create a symlink
    # MacOS/base_library.zip -> ../Resources/base_library.zip.
    rename(
        path('${freeze_dir}/Contents/MacOS/base_library.zip'),
        path('${freeze_dir}/Contents/Resources/base_library.zip')
    )
    symlink(
        '../Resources/base_library.zip',
        path('${freeze_dir}/Contents/MacOS/base_library.zip')
    )